
   /**------ ( ----------------------------------------------------------**
    **       )\                      CAnDL                               **
    **----- /  ) --------------------------------------------------------**
    **     ( * (                    space.h                              **
    **----  \#/  --------------------------------------------------------**
    **    .-"#'-.         First version: March 6th 2006                  **
    **--- |"-.-"| -------------------------------------------------------**
          |     | 
          |     | 
 ******** |     | *************************************************************
 * CAnDL  '-._,-' the Chunky Analyzer for Dependences in Loops (experimental) *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2006 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CAnDL, the Chunky Dependence Analyzer                                      *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 ******************************************************************************/


#ifndef CANDL_SPACE_H
#define CANDL_SPACE_H
#if defined(__cplusplus)
extern "C" 
  {
#endif 


/**
 * CandlSpace structure:
 * this structure contains all the informations about a legal transformation
 * space part.
 */
struct candlspace
{ int nb_statements ;        /**< Number of statements. */
  int nb_dependences ;       /**< Number of dependences. */
  int max_depth ;            /**< Maximum dependence depth (max of lexico). */

  /*int nb_parameters ;       **< Number of parameters. */
  /*int nb_unknowns ;         **< Number of unknowns. */
  /*Entier * delta ;          **< Delta. */

  int * lexico ;             /**< Lexicographic informations on the
                              *   dependences.  It is an array of
                              *   nb_dependences entries: lexico[i]
                              *   gives the level where the difference
                              *   of the transformations of the
                              *   statements involved in the ith
                              *   dependence is positive (thus it is
                              *   == 0 for every outer dimensions).
                              */
  int ** index ;             /**< Column index of each transformation
                              *   dimension in the domain matrix. It
                              *   is an array of nb_statements rows
                              *   and max_depth+1 columns: index[i][j]
                              *   gives the first column index of the
                              *   jth dimension of the transformation
                              *   function of the ith
                              *   statement. index[i][0] precises how
                              *   many dimensions have to be taken
                              *   into consideration for statement
                              *   i. The value of index[i][j] is
                              *   CANDL_UNSET if the dimension is not
                              *   considered.
                              */
  CandlMatrix * domain ;     /**< Legal transformation space domain. */
  struct candlspace * next ; /**< Next legal transformation space part. */
} ;
typedef struct candlspace CandlSpace ;


/******************************************************************************
 *                          Structure display function                        *
 ******************************************************************************/
void candl_space_print_structure(FILE *, CandlSpace *, int) ;
void candl_space_print(FILE *, CandlSpace *) ;
void candl_space_pprint(FILE *, CandlSpace *) ;


/******************************************************************************
 *                         Memory deallocation function                       *
 ******************************************************************************/
void candl_space_free(CandlSpace *) ;


/******************************************************************************
 *                             Processing functions                           *
 ******************************************************************************/
CandlSpace     * candl_space_malloc() ;
CandlSpace     * candl_space(CandlProgram *, CandlDependence *, CandlOptions *);


    /**********
     * Loulou *
     *********/

    struct {

      CandlStatement*		ref;
      CandlMatrix*		domain; /* NULL-terminated array of legal
				   transformation space */
      struct candl_space_s*	next;
    } candl_space_s;

    typedef struct candl_space_s candl_space_t;


    candl_space_t*	candl_build_ltspace(CandlDependence* dependences);
    


    

#if defined(__cplusplus)
  }
#endif 
#endif /* define _H */

