
   /**------ ( ----------------------------------------------------------**
    **       )\                      CAnDL                               **
    **----- /  ) --------------------------------------------------------**
    **     ( * (                   fourier.c                             **
    **----  \#/  --------------------------------------------------------**
    **    .-"#'-.         First version: April 3rd 2006                  **
    **--- |"-.-"| -------------------------------------------------------**
          |     |
          |     |
 ******** |     | *************************************************************
 * CAnDL  '-._,-' the Chunky Analyzer for Dependences in Loops (experimental) *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2006 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CAnDL, the Chunky Dependence Analyzer                                      *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 *****************************************************************************/

/* CAUTION: the english used for comments is probably the worst you ever read,
 *          please feel free to correct and improve it !
 */


# include <stdlib.h>
# include <stdio.h>
# include "../include/candl/candl.h"


static
int gcd_s(int x, int y)
{
  while (x !=y)
    if (x > y)
      x = x - y;
    else
      y = y - x;
  return x;
}

static 
int gcd(int a, int b)
{
  if (a < 0)
    a *= -1;
  if (b < 0)
    b *= -1;
  if (a < b)
    return gcd_s(a, b);
  else
    return gcd_s(b, a);
}

/*
 * Addition two rows.
 *
 */
void			candl_fourier_row_add(PipVector* res,
					      PipVector* val1,
					      PipVector* val2)
{
  unsigned		i, size = res->nb_elements;
  Entier		num, denom, tmp;

  // If mixing an equality and an inequality, it becomes an inequality.
  if ((CANDL_cmp_si(val1->the_vector[0], ==, 1)) ||
      (CANDL_cmp_si(val2->the_vector[0], ==, 1)))
    CANDL_set_si(res->the_vector[0], 1);
  else
    CANDL_set_si(res->the_vector[0], 0);

  // Add the rows, and simplify them.
  for (i = 1; i < size; ++i)
    {
      // val1[i] = 0 -> copy val2[i].
      if (CANDL_cmp_si(val1->the_vector[i], ==, 0))
	{
	  CANDL_assign(res->the_vector[i], val2->the_vector[i]);
	  CANDL_assign(res->the_deno[i], val2->the_deno[i]);
	}
      // val2[i] = 0 -> copy val1[i].
      else if (CANDL_cmp_si(val2->the_vector[i], ==, 0))
	{
	  CANDL_assign(res->the_vector[i], val1->the_vector[i]);
	  CANDL_assign(res->the_deno[i], val1->the_deno[i]);
	}
      else
	{
	  CANDL_multiply(num, val1->the_vector[i], val2->the_deno[i]);
	  CANDL_multiply(tmp, val2->the_vector[i], val1->the_deno[i]);
	  CANDL_addition(num, num, tmp);

	  // No need to worry if numerator is 0.
	  if (CANDL_cmp_si(num, ==, 0))
	    CANDL_set_si(denom, 1);
	  else
	    {
	      CANDL_multiply(denom, val1->the_deno[i], val2->the_deno[i]);
	      CANDL_gcd(tmp, num, denom);

	      // No need to simplify if gcd is 1.
	      if (CANDL_cmp_si(tmp, >, 1))
		{
		  CANDL_divide(num, num, tmp);
		  CANDL_divide(denom, denom, tmp);
		}
	    }
	  CANDL_assign(res->the_vector[i], num);
	  CANDL_assign(res->the_deno[i], denom);

	  // Simplify -n/-k in n/k.
	  if (CANDL_cmp_si(res->the_vector[i], <=, 0) &&
	      CANDL_cmp_si(res->the_deno[i], <, 0))
	    {
	      CANDL_oppose(res->the_vector[i], res->the_deno[i]);
	      CANDL_oppose(res->the_deno[i], res->the_deno[i]);
	    }
	}
    }
}



/*
 *
 *
 */
void			candl_fourier_row_assign(PipVector* row,
						 Entier* data)
{
  unsigned		i;

  for (i = 0; i < row->nb_elements; ++i)
    CANDL_assign(row->the_vector[i], data[i]);
}


/*
 *
 *
 */
void			candl_fourier_row_assign_v(PipVector* row1,
						   PipVector* row2)
{
  unsigned		i;

  for (i = 0; i < row2->nb_elements; ++i)
    {
      CANDL_assign(row1->the_vector[i], row2->the_vector[i]);
      CANDL_assign(row1->the_deno[i], row2->the_deno[i]);
    }
}


/*
 * Normalize the row (by the coef. of the given variable), and arrange
 * the (in-)equality.
 *
 */
void			candl_fourier_row_normalize(PipVector** data,
						    unsigned idx_var)
{
  unsigned		i;
  unsigned		size;
  Entier		tmp;

  if (data == NULL || data[0] == NULL)
    return;

  size = data[0]->nb_elements;

  while (*data)
    {
      if (CANDL_cmp_si((*data)->the_vector[idx_var], !=, 0))
	for (i = 1; i < size; ++i)
	  {
	    // Skip variables which are not in the (in-)equality.
	    if (CANDL_cmp_si((*data)->the_vector[i], ==, 0))
	      continue;

	    // Normalize with the coef. of the variable to remove.
	    CANDL_assign(tmp, (*data)->the_vector[idx_var]);
	    CANDL_abs(tmp, tmp);
	    CANDL_multiply((*data)->the_deno[i], (*data)->the_deno[i], tmp);
	    CANDL_gcd(tmp, (*data)->the_vector[i], (*data)->the_deno[i]);
	    if (CANDL_cmp_si((*data)->the_vector[i], !=, 1))
	      {
		CANDL_divide((*data)->the_deno[i], (*data)->the_deno[i], tmp);
		CANDL_divide((*data)->the_vector[i], (*data)->the_vector[i],
			     tmp);
	      }
	    if (CANDL_cmp_si((*data)->the_vector[i], <=, 0) &&
		CANDL_cmp_si((*data)->the_deno[i], <, 0))
	      {
		CANDL_oppose((*data)->the_vector[i], (*data)->the_deno[i]);
		CANDL_oppose((*data)->the_deno[i], (*data)->the_deno[i]);
	      }
	  }

      // Re-arrange the (in-)equality.
      if (CANDL_cmp_si((*data)->the_vector[idx_var], >, 0))
	for (i = 1; i < size; ++i)
	  if (i != idx_var)
	    CANDL_oppose((*data)->the_vector[i], (*data)->the_vector[i]);

      ++data;
    }
}


/*
 * Sort rows in the array, with (+) then (-) then (0) coef. of the
 * variable.
 *
 */
void			candl_fourier_row_sort(PipVector** data,
					       unsigned idx_var)
{
  unsigned		i, j;
  unsigned		count = 0;
  PipVector*		tmp;

  // Useful guard.
  if (data == NULL || data[0] == NULL)
    return;
    
  // Sort the (+) rows first.
  for (i = 0; data[i] != NULL; ++i)
    for (j = 0; data[j] != NULL; ++j)
      {
	if (CANDL_cmp_si(data[i]->the_vector[idx_var], >, 0) &&
	    CANDL_cmp_si(data[j]->the_vector[idx_var], <=, 0))
	  {
	    tmp = data[i];
	    data[i] = data[j];
	    data[j] = tmp;
	  }
      }

  // Count the number of (+) rows.
  for (i = 0; data[i] != NULL; ++i)
    if (CANDL_cmp_si(data[i]->the_vector[idx_var], >, 0))
      count++;

  // Sort the (-) rowe first, starting after the (+) rows.
  for (i = count; data[i] != NULL; ++i)
    for (j = count; data[j] != NULL; ++j)
      {
	if (CANDL_cmp_si(data[i]->the_vector[idx_var], <, 0) &&
	    CANDL_cmp_si(data[j]->the_vector[idx_var], ==, 0))
	  {
	    tmp = data[i];
	    data[i] = data[j];
	    data[j] = tmp;
	  }
      }
}



void			print_solution(PipVector** sol)
{
  if (sol == NULL)
    return;

  fprintf(stdout, "A solution: \n");

  while (*sol)
    {
      pip_vector_print(stdout, *sol);
      fprintf(stdout, "\n");
      sol++;
    }
}


/*
 *
 *
 */
PipVector**		candl_fourier_fill_solution(PipVector** system,
						    unsigned idx_var)
{
  PipVector**		solution = NULL;
  unsigned		n1 = 0, n2 = 0, size;
  unsigned		i, j;
  unsigned		count = 0;

  // Useful guard.
  if (system == NULL)
    return NULL;

  // Compute n1 and n2, indexes for the (+) rows and the (0) rows.
  for (i = 0, size = 0; system[i] != NULL; ++i, ++size)
    {
      if (CANDL_cmp_si(system[i]->the_vector[idx_var], >, 0))
	n1++;
      else if (CANDL_cmp_si(system[i]->the_vector[idx_var], <, 0))
	n2++;
    }
  n2 += n1;

  // Stop when no more (in-)equalities remains.
  if (size - n2 + n1 * (n2 - n1) <= 0)
    return NULL;

  // Allocate the new system.
  solution = (PipVector**) malloc((size - n2 + n1 * (n2 - n1) + 1) *
				  sizeof(PipVector*));

  // Array is NULL-terminated.
  for (i = 0; i < size - n2 + n1 * (n2 - n1) + 1; ++i)
    solution[i] = NULL;
/*   solution[size - n2 + n1 * (n2 - n1)] = NULL; */

  // Fill the new system.
  for (i = 0; i < n1; ++i)
    for (j = n1; j < n2; ++j)
      {
	solution[count] = pip_vector_alloc(system[i]->nb_elements);
	candl_fourier_row_add(solution[count], system[i], system[j]);
	++count;
      }

  // Add (in-)equalities which doesn't contain given variable.
  for (i = n2; i < size; ++i)
    {
      solution[count] = pip_vector_alloc(system[i]->nb_elements);
      candl_fourier_row_assign_v(solution[count], system[i]);
      ++count;
    }
  
  return solution;
}



/*
 *
 *
 */
PipVector*		pip_vector_alloc(const unsigned size)
{
  PipVector*		res = NULL;
  unsigned		i;

  if (((res = (PipVector*) malloc(sizeof(PipVector))) == NULL) ||
      ((res->the_vector = (Entier *) malloc(size * sizeof(Entier))) == NULL) ||
      ((res->the_deno = (Entier *) malloc(size * sizeof(Entier))) == NULL))
    {
      fprintf(stderr, "Memory Overflow.\n");
      exit(1);
    }

  for (i = 0; i < size; ++i)
    {
      CANDL_set_si(res->the_vector[i], 0);
      CANDL_set_si(res->the_deno[i], 1);
    }

  res->nb_elements = size;

  return res;
}


/*
 * Fourier-Motzkin elimination algorithm.
 *
 * @brief Fourier-Motzkin elimination algorithm, as proposed in
 * Barnejee's book.
 *
 * @input in A PipLib formated matrix, containing the system to solve.
 * @output res A PipLib formated matrix containing all the (integral)
 * constraints.
 *
 * @author Louis-Noel Pouchet <louis-noel.pouchet@inria.fr>
 *
 */
CandlMatrix*		candl_fourier_elimination(CandlMatrix* in)
{
  unsigned		dim = in->NbColumns;
  unsigned		size = in->NbRows;
  unsigned		i, count = 0;
  PipVector**		system;
  PipVector***		solution;

  // Useful guard.
  if (in == NULL || in->NbColumns < 2)
    return NULL;

  // Initialize the system.
  system = (PipVector**) malloc((size + 1) * sizeof(PipVector*));
  for (i = 0; i < size; ++i)
    {
      system[i] = pip_vector_alloc(dim);
      candl_fourier_row_assign(system[i], in->p[i]);
    }
  // Arrays are NULL-terminated.
  system[i] = NULL;

  // Initialize the solutions array, and do the first iteration of the
  // algorithm.
  solution = (PipVector***) malloc(dim * sizeof(PipVector**));
  candl_fourier_row_normalize(system, 1);
  candl_fourier_row_sort(system, 1);
  solution[0] = candl_fourier_fill_solution(system, 1);

  print_solution(solution[0]);
  fprintf(stdout, "dim: %d\n", dim);

  // Iterate the algorithm on the remaining variables.
  for (i = 2; i < dim; ++i)
    {
      candl_fourier_row_normalize(solution[i - 2], i);
      printf("after normalization: \n");
      print_solution(solution[i - 2]);
      candl_fourier_row_sort(solution[i - 2], i);
      printf("after sort: \n");
      print_solution(solution[i - 2]);
      solution[i - 1] = candl_fourier_fill_solution(solution[i - 2], i);
      printf("after fill: \n");
      print_solution(solution[i - 1]);
    }

  // temp...
  return in;
}
