
   /**------ ( ----------------------------------------------------------**
    **       )\                      CAnDL                               **
    **----- /  ) --------------------------------------------------------**
    **     ( * (                    space.c                              **
    **----  \#/  --------------------------------------------------------**
    **    .-"#'-.         First version: March 6th 2006                  **
    **--- |"-.-"| -------------------------------------------------------**
          |     |
          |     |
 ******** |     | *************************************************************
 * CAnDL  '-._,-' the Chunky Analyzer for Dependences in Loops (experimental) *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2006 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CAnDL, the Chunky Dependence Analyzer                                      *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 *****************************************************************************/

/* CAUTION: the english used for comments is probably the worst you ever read,
 *          please feel free to correct and improve it !
 */


# include <stdlib.h>
# include <stdio.h>
# include "../include/candl/candl.h"


/*****************************************************************************
 *                          Structure display function                       *
 *****************************************************************************/


/**
 * candl_space_print_structure function:
 * Displays a CandlSpace structure (space) into a file (file,
 * possibly stdout) in a way that trends to be understandable without falling
 * in a deep depression or, for the lucky ones, getting a headache... It
 * includes an indentation level (level) in order to work with others
 * print_structure functions.
 * - 06/03/2006: first version.
 */
void candl_space_print_structure(file, space, level)
FILE * file ;
CandlSpace * space ;
int level ;
{ int j, i, first=1 ;

  if (space != NULL)
  { /* Go to the right level. */
    for(j=0; j<level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"+-- CandlSpace\n") ;
  }
  else
  { for(j=0; j<level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"+-- NULL CandlSpace\n") ;
  }

  while (space != NULL)
  { if (!first)
    { /* Go to the right level. */
      for(j=0; j<level; j++)
      fprintf(file,"|\t") ;
      fprintf(file,"|   CandlSpace\n") ;
    }
    else
    first = 0 ;

    /* A blank line. */
    for(j=0; j<=level+1; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"\n") ;

    /* Go to the right level and print the statement number. */
    for(j=0; j<=level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"Nb statements: %d\n",space->nb_statements) ;


     /* A blank line. */
    for(j=0; j<=level+1; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"\n") ;

    /* Go to the right level and print the dependence number. */
    for(j=0; j<=level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"Nb dependences: %d\n",space->nb_dependences) ;

    /* A blank line. */
    for(j=0; j<=level+1; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"\n") ;

    /* Go to the right level and print the maximum depth. */
    for(j=0; j<=level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"Maximum depth: %d\n",space->max_depth) ;

    /* A blank line. */
    for(j=0; j<=level+1; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"\n") ;

    /* Go to the right level and print the space lexicographic informations. */
    for(j=0; j<=level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"+-- Dependence lexicographic informations:\n") ;

    for(j=0; j<=level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"\tNumber  |") ;
    for(j=0; j<space->nb_dependences; j++)
    fprintf(file,"%3d|",j) ;
    fprintf(file,"\n") ;

    for(j=0; j<=level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"\tDim > 0 |") ;
    for(j=0; j<space->nb_dependences; j++)
    fprintf(file,"%3d|",space->lexico[j]) ;
    fprintf(file,"\n") ;

    /* A blank line. */
    for(j=0; j<=level+1; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"\n") ;

    /* Go to the right level and print the dimension index informations. */
    for(j=0; j<=level; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"+-- Index of transformation dimensions in domain:\n") ;

    for(j=0; j<=level+1; j++)
    fprintf(file,"|\t") ;
    fprintf(file,"Label   |") ;
    for(j=0; j<space->nb_statements; j++)
    fprintf(file,"%3d|",j) ;
    fprintf(file,"\n") ;

    for(i=0; i<space->max_depth; i++)
    { for(j=0; j<=level+1; j++)
      fprintf(file,"|\t") ;

      fprintf(file,"Dim %3d |",i) ;

      for(j=0; j<space->nb_statements; j++)
      { if (space->index[j][i] != CANDL_UNSET)
        fprintf(file,"%3d|",space->index[j][i]) ;
        else
        fprintf(file," # |") ;
      }
      fprintf(file,"\n") ;
    }

    /* Print the legal transformation part polyhedron. */
    candl_matrix_print_structure(file,space->domain,level+1) ;

    space = space->next ;

    /* Next line. */
    if (space != NULL)
    { for (j=0; j<=level; j++)
      fprintf(file,"|\t") ;
      fprintf(file,"V\n") ;
    }
  }

  /* The last line. */
  for(j=0; j<=level; j++)
  fprintf(file,"|\t") ;
  fprintf(file,"\n") ;
}


/* candl_space_print function:
 * This function prints the content of a CandlSpace structure (space) into a
 * file (file, possibly stdout).
 */
void candl_space_print(FILE * file, CandlSpace * space)
{ candl_space_print_structure(file,space,0) ;
}


/******************************************************************************
 *                         Memory deallocation function                       *
 ******************************************************************************/


/* candl_space_free function:
 * This function frees the allocated memory for a CandlSpace structure.
 */
void candl_space_free(CandlSpace * space)
{ int i ;
  CandlSpace * next ;

  while (space != NULL)
  { next = space->next ;

    for (i=0;i<space->nb_dependences;i++)
    free(space->index[i]) ;
    free(space->index) ;
    free(space->lexico) ;
    candl_matrix_free(space->domain) ;
    free(space) ;

    space = next ;
  }
}


/******************************************************************************
 *                            Processing functions                            *
 ******************************************************************************/


/**
 * candl_space_malloc function:
 * This function allocates the memory space for a CandlSpace structure and
 * sets its fields with default values. Then it returns a pointer to the
 * allocated space.
 * - 06/03/2006: first version.
 */
CandlSpace * candl_space_malloc()
{ CandlSpace * space ;

  /* Memory allocation for the CandlSpace structure. */
  space = (CandlSpace *)malloc(sizeof(CandlSpace)) ;
  if (space == NULL)
  { fprintf(stderr, "[Candl]ERROR: memory overflow.\n") ;
    exit(1) ;
  }

  /* We set the various fields with default values. */
  space->nb_statements  = 0 ;
  space->nb_dependences = 0 ;
  space->max_depth      = 0 ;
  space->lexico         = NULL ;
  space->index          = NULL ;
  space->domain         = NULL ;
  space->next           = NULL ;

  return space ;
}


/**
 * candl_space_add function:
 * This function adds a CandlSpace structure (space) at a given place
 * (now) of a NULL terminated list of CandlSpace structures. The beginning
 * of this list is (start). This function updates (now) to the end of the loop
 * list (loop), and updates (start) if the added element is the first one -that
 * is when (start) is NULL-.
 * - 18/09/2003: first version.
 */
void candl_space_add(start, now, space)
CandlDependence ** start, ** now, * space ;
{ if (space != NULL)
  { if (*start == NULL)
    { *start = space ;
      *now = *start ;
    }
    else
    { (*now)->next = space ;
      *now = (*now)->next ;
    }

    while ((*now)->next != NULL)
    *now = (*now)->next ;
  }
}


CandlSpace * candl_space_part(program, dependence, nb_dependences, lexi,options)
CandlProgram * program ;
CandlDependence * dependence ;
int nb_dependences, * lexi ;
CandlOptions * options ;
{ int i, j, nb_unknowns, nb_inequalities, column, max_depth ;
  int * lexico, ** index ;
  CandlDependence * temp ;
  /* CandlMatrix * domain ; */
  CandlSpace * space ;

  /* We compute max_depth : the maximum value amongst lexi. */
  max_depth = 0 ;
  for (i=0; i<nb_dependences; i++)
  if (lexi[i] > max_depth) ;
  max_depth = lexi[i] ;

  /* We allocate and fill lexico array, a copy of the lexi array to be
   * included inside the CandlSpace structure.
   */
  lexico = (int *)malloc(nb_dependences*sizeof(int)) ;
  for (i=0; i<nb_dependences; i++)
  lexico[i] = lexi[i] ;

  /* We allocate and fill the index array: */
  /* - Allocation. */
  index = (int **)malloc(program->nb_statements*sizeof(int)) ;
  for (i=0; i<program->nb_statements; i++)
  index[i] = (int *)malloc((max_depth+1)*sizeof(int)) ;

  /* - Initialization to CANDL_UNSET. */
  for (i=0; i<program->nb_statements; i++)
  for (j=0; j<=max_depth; j++)
  index[i][j] = CANDL_UNSET ;

  /* - Filling of the index array. */
  /*   + First of all we have to determine how many dimensions are involved for
   *     each statements: this is the maximum of the lexi entries where the
   *     statement is concerned. We also begin to compute the total number of
   *     unknowns of the system.
   */
  nb_unknowns = 0 ;
  temp = dependence ;
  while (dependence != NULL)
  { /* The number of Farkas multipliers is the number of INequalities in the
     * statement domain plus one (see Bastoul's thesis).
     */
    nb_inequalities = 0 ;
    for (i=0; i<dependence->domain->NbRows; i++)
    nb_inequalities += (dependence->domain->p[i][0] == 0) ? 2 : 1 ;
    nb_unknowns += nb_inequalities + 1 ;

    if (lexi[dependence->source->label] > index[dependence->source->label][0])
    index[dependence->source->label][0] = lexi[dependence->source->label] ;

    if (lexi[dependence->target->label] > index[dependence->target->label][0])
    index[dependence->target->label][0] = lexi[dependence->target->label] ;

    dependence = dependence->next ;
  }
  dependence = temp ;

  /*   + Next we fill the index array. We also compute the total number of
   *     unknowns of the system.
   */
  column = 1 ;
  for (i=0; i<program->nb_statements; i++)
  { if (index[i][0] > 0)
    { /* The number of Farkas multipliers is the number of INequalities in the
       * statement domain plus one (see Bastoul's thesis).
       */
      nb_inequalities = 0 ;
      for (j=0; j<program->statement[i]->domain->NbRows; j++)
      nb_inequalities += (program->statement[i]->domain->p[j][0] == 0) ? 2 : 1 ;
      nb_unknowns += nb_inequalities + 1 ;

      for (j=1; j<=index[i][0]; j++)
      { index[i][j] = column ;
        column += program->statement[i]->domain->NbRows + 1 ;
      }
    }
  }

  /* Memory allocation for the CandlSpace structure. */
  space = candl_space_malloc() ;

  space->nb_statements  = program->nb_statements ;
  space->nb_dependences = nb_dependences ;
  space->max_depth      = max_depth ;
  space->lexico         = lexico ;
  space->index          = index ;

  return space ;
}







CandlSpace * candl_space(program, dependence, options)
CandlProgram * program ;
CandlDependence * dependence ;
CandlOptions * options ;
{ int  nb_dependences, max_depth, depth ;
  int i, j, col_s, col_t, * lexi, k, l ;
  CandlDependence * temp ;

  /* We make a traversal of the dependences to find the dependence number
   * and the maximum transformation function dimension according to dependences.
   */
  max_depth = 0 ;
  nb_dependences = 0 ;
  temp  = dependence ;
  while (dependence != NULL)
  { nb_dependences ++ ;
    col_s = program->statement[dependence->source->label]->domain->NbColumns ;
    col_t = program->statement[dependence->target->label]->domain->NbColumns ;

    depth = (col_s > col_t) ? col_t : col_s ;
    depth -= program->context->NbColumns ;

    if (depth > max_depth)
    max_depth = depth ;

    dependence = dependence->next ;
  }
  dependence = temp ;


  lexi = (int *)malloc(nb_dependences*sizeof(int)) ;
  for (i=0; i<nb_dependences; i++)
  lexi[i] = 0 ;

  /* Et faire des break avant les tests quand on depasse la valeur max. */
  for (l=0; l<max_depth; l++)
  for (i=0; i<nb_dependences; i++)
  { if (i > 0)
    lexi[i-1] ++ ;
    else
    { /* Test */
      for (k=0; k<nb_dependences; k++)
      printf(" %3d |",lexi[k]) ;
      printf("\n") ;
    }

    for (j=i; j<nb_dependences; j++)
    { lexi[j] ++ ;

      /* Test */
      for (k=0; k<nb_dependences; k++)
      printf(" %3d |",lexi[k]) ;
      printf("\n") ;

      /* Don't remove it if this is the end of a i loop. */
      if ((i != nb_dependences-1) || (j != nb_dependences-1))
      lexi[j] -- ;
    }
  }


  return NULL ;
}
