#!/bin/sh
## installer.sh for Pouchet in ~/projects
## 
## Made by Loulou Pouchet
## Contact: <loulou@lrde.epita.fr>
## 
## Started on  Wed Nov 15 23:04:19 2006 Loulou
## Last update Sun Feb 17 18:34:53 2008 Louis-Noel Pouchet
##


POLYLIB_NAME="polylib-5.22.3"
PIPLIB_NAME="piplib-1.3.6-patched"
CANDL_NAME="candl-0.1.0-prerelease"
CLOOG_NAME="cloog-0.14.0-sv"
FM_NAME="fm-0.4.0"
LETSEE_NAME="letsee-0.1.0"


if [ $# -gt 1 ]; then 
    echo "Usage: ./installer.sh";
fi

## Set the installation prefix, if supplied in argument
if ! [ -z "$1" ]; then
  prefix="--prefix=$1";
  value="$1"
else
## Creating the installation directory.
    currentdir=`pwd`;
    prefix="$currentdir/install";
    mkdir -p $prefix;
    value="$prefix";
    prefix="--prefix=$prefix";
fi

## Set the MAKE environmend variable, if not supplied in the
## environment.
if [ -z "$MAKE" ]; then
    MAKE=make;
fi


function check_value
{
    if [ $1 -ne 0 ]; then 
	cd ..
	echo -e "\033[31m==> ERROR\033[0m"
	exit $1;
    fi
}

## Installing polylib.
echo -e "\033[31m* Installing Polylib\033[0m"
tar xzf $POLYLIB_NAME.tar.gz
cd $POLYLIB_NAME && ./configure $prefix && $MAKE && $MAKE install;
check_value $?
echo -e "\033[32m* Polylib Successfully installed\033[0m"

## Installing piplib.
echo -e "\033[31m* Installing PIPLib\033[0m"
cd .. && \
tar xzf $PIPLIB_NAME.tar.gz && \
cd $PIPLIB_NAME && ./configure $prefix && $MAKE && $MAKE install;
check_value $?
echo -e "\033[32m* PIPLib Successfully installed\033[0m"

## Installing cloog.
echo -e "\033[31m* Installing CLooG\033[0m"
cd .. && \
tar xzf $CLOOG_NAME.tar.gz && \
cd $CLOOG_NAME && ./configure $prefix --with-polylib-prefix=$value && $MAKE \
    && $MAKE install;
check_value $?
echo -e "\033[32m* CLooG successfully installed\033[0m"

## Installing Fourier-Motzkin.
echo -e "\033[31m* Installing FM\033[0m"
cd .. && \
tar xzf $FM_NAME.tar.gz && \
cd $FM_NAME && ./configure $prefix && $MAKE && $MAKE check && $MAKE install;
check_value $?
echo -e "\033[32m* FM successfully installed\033[0m"

## Installing candl.
echo -e "\033[31m* Installing Candl\033[0m"
cd .. && \
tar xzf $CANDL_NAME.tar.gz && \
cd $CANDL_NAME && ./configure $prefix --with-piplib=$value \
    --with-polylib=$value  && $MAKE && $MAKE install;
check_value $?
echo -e "\033[32m* Candl successfully installed\033[0m"

## Installing letsee.
echo -e "\033[31m* Installing LestSee\033[0m"
cd .. && \
tar xzf $LETSEE_NAME.tar.gz && \
cd $LETSEE_NAME && ./configure $prefix  && $MAKE;
check_value $?
echo -e "\033[32m=> LetSee successfully compiled\033[0m"
echo -e "\033[32m* See README file for further instructions.\033[0m"

curpath=`cd .. && pwd`
echo -e "\033[32mYou're encouraged to set the following environment variable:\033[0m"
echo -e "\033[32mLD_LIBRARY_PATH=$curpath/install/lib:\$LD_LIBRARY_PATH\033[0m"
echo -e "\033[32mPATH=$curpath/install/bin:\$PATH\033[0m"
